#include "block.h"
#include <assert.h>

/*** Utility functions ***/

/*
 * Used to make a file grow with 0s as data in new space
 */
static int grow(int fd, int oldsize, int newsize)
{
    if ( lseek(fd, oldsize, SEEK_SET) == (off_t)-1 ) {
	perror("block: lseek");
	return 0;
    }

    char * tab = calloc(newsize-oldsize, 1);

    if ( write(fd, tab, newsize-oldsize) != newsize-oldsize ) {
	perror("block: write");
	return 0;
    }

    free(tab);

    return 1;
}

/*
 * Find a free block in this superblock and set it as used
 */
static int get_free_block(super_block * superblock)
{
    int i, j;

    if ( superblock->used_blocks[0] & 1 ) // superblock full
	return -1;

    // For each byte
    for ( i = 0 ; i < sizeof(data_block) ; i++ ) {
	if ( superblock->used_blocks[i] != 255 )
	    // Look for a free bit
	    for ( j = ( i == 0 ) ? 1 : 0 ; j < CHAR_BIT ; j++ )
		if ( ! (superblock->used_blocks[i] & ( 1 << j ) ) ) {
		    // Found, so use it
		    superblock->used_blocks[i] |= ( 1 << j );
		    return i * CHAR_BIT + j - 1;
		}
    }

    superblock->used_blocks[0] = 255; // the superblock is full

    return -1;
}

/*
 * Get the superblock where block n is
 */
static inline super_block * get_superblock_for_n(block * blocksys, int n)
{
    return blocksys->superblocks + ( n / NBBLOCKS_IN_SUPERBLOCK );
}

/*
 * Return 1 if block n is in use in this superblock, 0 otherwise.
 */
static int blockinuse(super_block * sbl, int n)
{
    if ( sbl->used_blocks[0] == 255 ) // is superblock full ?
	return 1;

    n++; // first bit is block full => +1

    assert(n > 0 && n <= NBBLOCKS_IN_SUPERBLOCK);

    if ( sbl->used_blocks[n / CHAR_BIT] & ( 1 << (n % (CHAR_BIT)) ) )
	return 1;

    return 0;
}

/*** Actual interface ***/

/*
 * Init block system and check filesize
 */
block * block_init(int fd, off_t offset)
{
    int pagesize = getpagesize();
    block * blocksys = NULL;

    // Init block system
    blocksys = malloc(sizeof(block));
    if ( blocksys == NULL ) {
	perror("block: malloc");
	return NULL;
    }

    blocksys->fd = fd;
    blocksys->offset = offset;
    blocksys->gotomapstart = offset % pagesize;
    offset -= blocksys->gotomapstart;

    // Map filesystem beginning to memory
    blocksys->superblocks = mmap(NULL, MAP_SIZE | MAP_NORESERVE,
	    PROT_READ | PROT_WRITE, MAP_SHARED, fd, offset);
    if ( blocksys->superblocks == MAP_FAILED ) {
	perror("block: mmap");
	free(blocksys);
	return NULL;
    }
    blocksys->superblocks = (super_block *)
	((unsigned char *) blocksys->superblocks + blocksys->gotomapstart);
    blocksys->firstblock = 0;

    // Format if necessary
    struct stat buf;
    if ( fstat(fd, & buf) == -1 ) {
	perror("block: fstat");
	fprintf(stderr, "The program will continue but the filesystem will be "
		"erased.\n");
	blocksys->filesize = 0;
    } else
	blocksys->filesize = buf.st_size;

    return blocksys;
}

/*
 * Free block system memory
 */
void block_destroy(block * blocksys)
{
    blocksys->superblocks = (super_block *) 
	((unsigned char *) blocksys->superblocks - blocksys->gotomapstart);
    if ( msync(blocksys->superblocks, MAP_SIZE, MS_SYNC) == -1 )
	perror("block: msync");
    if ( munmap(blocksys->superblocks, MAP_SIZE) == -1 )
	perror("block: munmap");

    free(blocksys);
}

/*
 * Get a new fresh block
 */
int block_alloc(block * blocksys)
{
    static data_block zaroo;
    int i, size, n;
    data_block * data;

    // TODO verify the first mapped block and offset n if necessary
    // or remap if necessary

    // Look in each mapped superblock for a free block
    for ( i = 0 ; i < MAP_SIZE / sizeof(super_block) ; i++ ) {
	size = blocksys->offset + (i+1) * sizeof(super_block);
	if ( blocksys->filesize < size ) {
	    grow(blocksys->fd, blocksys->filesize, size);
	    blocksys->filesize = size;
	}
	if ( (n = get_free_block(blocksys->superblocks + i) ) != -1 ) {
	    n = n + (i * NBBLOCKS_IN_SUPERBLOCK);
	    // Clean the new block
	    data = block_get(blocksys, n);
	    if ( data == NULL )
		return -1;
	    memcpy(data, & zaroo, sizeof(data_block));
	    return n;
	}
    }

    return -1;
}

/*
 * Free one block
 */
void block_free(block * blocksys, int n)
{
    // TODO verify the first mapped block and offset n if necessary
    // or remap if necessary
    
    super_block * superblock = get_superblock_for_n(blocksys, n);
    n %= NBBLOCKS_IN_SUPERBLOCK; // n -> n in its superblock
    n++;
    superblock->used_blocks[n / CHAR_BIT] &= ~(1 << (n % CHAR_BIT));
}

/*
 * Get a pointer to block data
 */
data_block * block_get(block * blocksys, int n)
{
    super_block * superblock = get_superblock_for_n(blocksys, n);
    n %= NBBLOCKS_IN_SUPERBLOCK; // n -> n in its superblock

    if ( ! blockinuse(superblock, n) )
	return NULL;

    return superblock->data + n;
}
