#ifndef _BLOCK_H
#define _BLOCK_H

/*
 * This module permits to create/delete/read/write each block
 * of the file system.
 */

#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define MAP_SIZE 1073741824 // 1GB
#define NBBLOCKS_IN_SUPERBLOCK (sizeof(data_block)*CHAR_BIT-1)

/*
 * One block, 512 bytes.
 */
typedef unsigned char data_block[512];

/*
 * A super block.
 * It's one bit-block to store free/in use information of the data blocks
 * contained (the first bit tells if the super block is totally used or not)
 * and NBBLOCKS_IN_SUPERBLOCK data blocks.
 */
typedef struct {
    data_block used_blocks; // bit array for free/used blocks
    data_block data[NBBLOCKS_IN_SUPERBLOCK]; // data blocks
} super_block;

/*
 * This structure olds the needed data to make the block system work.
 * It permits to re-map the file at another offset to be able to write more data
 * than the mapped data.
 */
typedef struct {
    int fd; // file containing the block system
    off_t offset; // offset for the first block in the file
    size_t filesize; // file size to avoid calls to stat
    super_block * superblocks; // mapped to the file.
    off_t gotomapstart; // number to be substracted of superblocks to be able
			// to unmap it correctly as mmap requires an offset
			// multiple of getpagesize()
    off_t firstblock; // number of the superblock mapped at superblocks
} block;

/*
 * Init and destroy block system
 */
block * block_init(int fd, off_t offset);
void block_destroy(block * blocksys);

/*
 * Used to get free blocks and free them again
 *
 * Parameter Block must be created by a call to block_init.
 */
int block_alloc(block * Block);
void block_free(block * Block, int n);

/*
 * Used to read/write data to blocks
 *
 * Parameter n is given by a precedent call to block_alloc.
 */
data_block * block_get(block * Block, int n);

#endif /* _BLOCK_H */
