#include "dir.h"

#include <assert.h>

/*** Utilities ***/

/*
 * Make a new dir_entry
 */
static dir_entry * dir_entry_make(int inode, char * filename)
{
    dir_entry * de;

    de = malloc(sizeof(dir_entry));
    if ( de == NULL )
	return NULL;

    de->inode = inode;

    de->filename = malloc(strlen(filename) + 1);
    if ( de->filename == NULL ) {
	free(de);
	return NULL;
    }
    strcpy(de->filename, filename);

    return de;
}

/*
 * Free one dir_entry
 */
static void dir_entry_free(dir_entry * de)
{
    free(de->filename);
    free(de);
}

/*
 * Converts serial file data to dir_entries
 */
static int make_entries(dir * d, char * data)
{
    int inode, i;

    d->entries = malloc(sizeof(dir_entry *) * (* (int *) data));
    if ( d->entries == NULL )
	return 0;

    d->nb_entries = * (int *) data;
    data += sizeof(int);

    for ( i = 0 ; i < d->nb_entries ; i++ ) {
	inode = * (int *) data;
	data += sizeof(int);
	d->entries[i] = dir_entry_make(inode, data);
	if ( d->entries[i] == NULL ) {
	    while ( --i != -1 )
		dir_entry_free(d->entries[i]);
	    d->nb_entries = 0;
	    return 0;
	}
	data += strlen(data) + 1;
    }

    return 1;
}

/*
 * Converts dir_entries to serial file data
 */
static int make_data(dir * d, char ** data)
{
    int size = 0, i, len;

    *data = malloc(sizeof(int));
    if ( *data == NULL )
	return 0;

    memcpy(*data, & (d->nb_entries), sizeof(int));
    size = sizeof(int);

    for ( i = 0 ; i < d->nb_entries ; i++ ) {
	len = strlen(d->entries[i]->filename) + 1;

	*data = realloc(*data, size + len + sizeof(int));
	// copy inode
	memcpy(*data + size, & (d->entries[i]->inode), sizeof(int));
	// copy filename
	strcpy(*data + size + sizeof(int), d->entries[i]->filename);

	size += sizeof(int) + len;
    }

    return size;
}

/*
 * Search for one name between directory entries
 * Returns 1 if found, 0 otherwise.
 */
static int search(dir * d, char * name)
{
    int i;
    for ( i = 0 ; i < d->nb_entries ; i++ )
	if ( strcmp(d->entries[i]->filename, name) == 0 )
	    return 1;
    return 0;
}

/*
 * Store directory data to its file
 */
static int apply(dir * d)
{
    char * data = NULL;
    int datasize;

    datasize = make_data(d, & data);

    if ( data == NULL )
	return -1;

    file f = { d->ino->address, d->ino->size };

    if ( file_set(d->fs, &f, 0, datasize, data) != datasize ) {
	free(data);
	return 0;
    }

    // Free some blocks not used anymore if an entry was deleted
    if ( file_truncate(d->fs, &f, datasize) == 0 )
	fprintf(stderr, "dir: file_truncate failed\n");

    d->ino->size = datasize;

    free(data);

    return 1;
}

/*** Implementation ***/

/*
 * Init directory filesystem
 * This formats the filesystem if necessary by adding a root directory
 */
int dir_init(filesys * fs, inodesys * is)
{
    struct stat st;
    dir * d;
    inode * ino;

    ino = inode_get(is, 0);
    if ( ino == NULL ) {
	// Create root directory
	d = dir_create(fs, is);
	if ( d == NULL || ! S_ISDIR(d->ino->mode) ) {
	    fprintf(stderr, "dir: could not create root directory !\n");
	    return 0;
	}
	dir_free(d);
	fstat(is->fd, &st);
	ino = inode_get(is, 0);
	if ( ino == NULL ) {
	    fprintf(stderr, "Could not set root inode !\n");
	    return 0;
	}
	ino->mode |= st.st_mode & 0777;
	ino->access = ino->modification = ino->changestate = time(NULL);
	ino->uid = getuid();
	ino->gid = getgid();
    }

    if ( ! S_ISDIR(ino->mode) ) {
	fprintf(stderr, "dir: there is no root directory\n"
		"dir: this is not a L3 file system or "
		"file system is corrupted\n");
	return 0;
    }

    return 1;
}

/*
 * Return a new directory
 */
dir * dir_create(filesys * fs, inodesys * is)
{
    dir * d;
    
    d = malloc(sizeof(dir));

    if ( d == NULL )
	return NULL;

    d->inodenb = inode_alloc(is);
    if ( d->inodenb == -1 ) {
	free(d);
	return NULL;
    }

    if ( (d->ino = inode_get(is, d->inodenb)) == NULL ) {
	inode_free(is, d->inodenb);
	free(d);
	return NULL;
    }

    file * f = file_alloc(fs);
    if ( f == NULL ) {
	inode_free(is, d->inodenb);
	free(d);
	return NULL;
    }

    d->fs = fs;
    d->is = is;
    d->ino->mode = S_IFDIR;
    d->ino->nlink = 2; // "." & filename
    d->ino->size = 0;
    d->ino->address = f->faddrblck;
    d->nb_entries = 0;
    d->entries = malloc(0);

    if ( ! apply(d) ) {
	file f = { d->ino->address, d->ino->size };
	file_free(d->fs, & f);
	dir_free(d);
    }

    free(f);

    return d;
}

/*
 * Return directory pointed by dirinode
 */
dir * dir_get(filesys * fs, inodesys * is, int dirinode)
{
    inode * ino;
    file f;
    dir * d;
    char * dirdata;

    if ( (ino = inode_get(is, dirinode)) == NULL || ! S_ISDIR(ino->mode) )
	return NULL;

    d = malloc(sizeof(dir));
    if ( d == NULL )
	return NULL;

    f.size = ino->size;
    f.faddrblck = ino->address;

    dirdata = malloc(f.size);
    if ( dirdata == NULL ) {
	free(d);
	return NULL;
    }

    if ( file_get(fs, & f, 0, f.size, dirdata) != f.size ) {
	free(d);
	free(dirdata);
	return NULL;
    }
    
    d->fs = fs;
    d->is = is;
    d->inodenb = dirinode;
    d->ino = ino;
    make_entries(d, dirdata);

    free(dirdata);

    return d;
}

/*
 * Add an entry to the directory, this do not add any file/inode data, this only
 * adds the name of the file to the list of filenames of the directory
 */
int dir_add(dir * d, int inode, char * filename)
{
    if ( search(d, filename) )
	return -1;

    assert(strchr(filename, '/') == NULL);
    assert(strcmp(filename, ".") != 0);
    assert(strcmp(filename, "..") != 0);

    d->nb_entries++;
    d->entries = realloc(d->entries, sizeof(dir_entry *) * d->nb_entries);
    if ( d->entries == NULL )
	return 0;
    d->entries[d->nb_entries - 1] = dir_entry_make(inode, filename);

    if ( ! apply(d) )
	return 0;

    return 1;
}

/*
 * Remove an entry from directory, this do not remove the file, it only
 * removes its name from the directory
 */
int dir_remove(dir * d, char * filename)
{
    int i, del = 0;

    // Search file to remove and copy others
    for ( i = 0 ; i < d->nb_entries ; i++ ) {
	if ( strcmp(d->entries[i]->filename, filename) != 0 ) {
	    if ( del ) {
		dir_entry_free(d->entries[i-del]);
		d->entries[i-del] = dir_entry_make(d->entries[i]->inode,
			d->entries[i]->filename);
	    }
	    if ( d->entries[i-del] == NULL )
		return 0;
	} else {
#ifdef NDEBUG
	    del = 1;
#else
	    del++;
	    assert(del < 2);
#endif
	}
    }

    if ( ! apply(d) )
	return 0;

    return 1;
}

/*
 * Delete given directory from filesystem
 * The directory must be emptied before !
 */
int dir_delete(dir * d)
{
    file f;

    assert(d->nb_entries == 0);
    assert(d->inodenb > 0); // Please do not remove / oh damn don't...

    f.faddrblck = d->ino->address;
    f.size = d->ino->size;
    if ( ! file_free(d->fs, & f) )
	return 0;
    inode_free(d->is, d->inodenb);

    return 1;
}

/*
 * Free memory allocated for given dir structure
 */
void dir_free(dir * d)
{
    int i;
    for ( i = 0 ; i < d->nb_entries ; i++ )
	dir_entry_free(d->entries[i]);
    free(d->entries);
    free(d);
}
