#include "file.h"

#include <assert.h>

#define SIZEOFADDRBLOCK (sizeof(data_block) / sizeof(off_t))
#define BLOCKSINADDR (sizeof(data_block) / sizeof(off_t) - 1)
#define NEXTADDR (sizeof(data_block) / sizeof(off_t) - 1)

/*** Utilities ***/

/*
 * Returns the block number corresponding to the given file offset
 */
static off_t get_blocknmbrbyoffset(filesys * fs, file * f, off_t offset)
{
    int pass, i;
    off_t (* address)[SIZEOFADDRBLOCK];
    off_t curaddr = f->faddrblck;

    assert(offset <= f->size);

    offset /= sizeof(data_block); // Convert offset to blocks instead of bytes

    pass = offset / BLOCKSINADDR; // number of address blocks to pass

    if ( ! (address = (off_t (*)[]) block_get(fs->blocksys, curaddr)) )
	return 0;

    for ( i = 0 ; i < pass ; i++ ) {
	curaddr = (* address)[NEXTADDR];
	offset -= BLOCKSINADDR;
	if ( (address = (off_t (*)[]) block_get(fs->blocksys, curaddr)) == NULL)
	    return 0;
    }

    assert((* address)[offset] != 0);

    return (* address)[offset];
}

/*
 * Returns the address block number refering the given file offset
 */
static off_t get_addrblcknbrbyoffset(filesys * fs, file * f, off_t offset)
{
    int pass, i;
    off_t (* address)[SIZEOFADDRBLOCK];
    off_t curaddr = f->faddrblck;

    assert(offset <= f->size);

    // number of address blocks to pass
    pass = offset / sizeof(data_block) / BLOCKSINADDR;

    if ( ! (address = (off_t (*)[]) block_get(fs->blocksys, curaddr)) )
	return 0;

    for ( i = 0 ; i < pass ; i++ ) {
	curaddr = (* address)[NEXTADDR];
	if ( (address = (off_t (*)[]) block_get(fs->blocksys, curaddr)) == NULL)
	    return 0;
    }

    assert(curaddr != 0);

    return curaddr;
}

/*
 * Make a file grow to a new size
 */
static size_t grow(filesys * fs, file * f, size_t newsize)
{
    size_t i;
    off_t (* address)[SIZEOFADDRBLOCK],
	  curaddr;

    // Add the rest of the last block to the filesize
    if ( (i = f->size % sizeof(data_block)) )
	f->size += sizeof(data_block) - i;

    // Get last address block
    curaddr = get_addrblcknbrbyoffset(fs, f, f->size);
    if ( curaddr == 0 )
	return f->size;

    // Add blocks if necessary
    while ( f->size <= newsize ) {
	if ( (address = (off_t (*)[]) block_get(fs->blocksys, curaddr)) == NULL)
	    return -1;

	// Add data blocks
	for ( i = 0 ; i < BLOCKSINADDR && f->size <= newsize ; i++ ) {
	    if ( (* address)[i] == 0 ) {
		(* address)[i] = block_alloc(fs->blocksys);
		if ( (*address)[i] == 0 )
		    return -1;
		f->size += sizeof(data_block);
	    }
	}

	// Add address block
	if ( f->size <= newsize ) {
	    curaddr = block_alloc(fs->blocksys); 
	    if ( curaddr == 0 )
		return -1;
	    (* address)[NEXTADDR] = curaddr;
	}
    }

    // Set correct size, as actual size is only a multiple of sizeof(data_block)
    f->size = newsize;

    return f->size;
}

/*
 * Shrink a file by freeing blocks
 */
static int shrink(filesys * fs, file * f, size_t newsize)
{
    size_t i, oldsize = f->size;
    off_t (* address)[SIZEOFADDRBLOCK],
	  curaddr, blck;
    
    // Free every block > block where newsize is
    for ( i = (newsize % sizeof(data_block) == 0) ? newsize :
	    newsize + sizeof(data_block) - newsize % sizeof(data_block)
	    ; i < oldsize ; i += sizeof(data_block) ) {
	curaddr = get_addrblcknbrbyoffset(fs, f, i);
	if ( curaddr == 0 )
	    return f->size;

	address = (off_t (*)[]) block_get(fs->blocksys, curaddr);
	if ( address == NULL )
	    return f->size;

	blck = (i / sizeof(data_block)) % BLOCKSINADDR;

	assert((* address)[blck] != 0);
	
	block_free(fs->blocksys, (* address)[blck]);

	if ( f->size > sizeof(data_block) )
	    f->size -= sizeof(data_block);
	else
	    f->size = 0;
    }

    assert(f->size >= newsize && f->size < newsize + sizeof(data_block));

    f->size = newsize;

    return f->size;
}

/*** Interface implementation ***/

/*
 * Initializes filesystem
 * (this formates the filesystem if necessary)
 */
filesys * file_init(int fd, off_t offset)
{
    int zaroO;
    filesys * fs = malloc(sizeof(filesys));

    if ( fs == NULL )
	return 0;

    fs->blocksys = block_init(fd, offset);
    if ( fs->blocksys == NULL ) {
	free(fs);
	return NULL;
    }

    /*
     * Format file system if necessary
     * ( by invalidating block zero )
     */
    zaroO = block_alloc(fs->blocksys);
    if ( zaroO > 0 ) // filesystem was already formatted
	block_free(fs->blocksys, zaroO);
    else if ( zaroO < 0 ) {
	free(fs);
	return NULL;
    }

    return fs;
}

/*
 * Free filesystem memory
 */
void file_destroy(filesys * fs)
{
    block_destroy(fs->blocksys);
    free(fs);
}

/*
 * Create a void file
 * ( allocates one empty address block )
 */
file * file_alloc(filesys * fs)
{
    file * f;
    
    f = malloc(sizeof(file));

    // Get starting block
    f->faddrblck = block_alloc(fs->blocksys);
    assert(f->faddrblck != 0);
    if ( f->faddrblck == -1 ) {
	free(f);
	return NULL;
    }

    f->size = 0;

    return f;
}

/*
 * Delete file data
 */
int file_free(filesys * fs, file * f)
{
    int i;
    off_t curaddr = f->faddrblck;
    off_t (* address)[SIZEOFADDRBLOCK];

    do {
	// Get next address block then free it
	if ( ! (address = (off_t (*)[])block_get(fs->blocksys, curaddr)) )
	    // File is corrupted
	    return 0;
	block_free(fs->blocksys, curaddr);
	
	// Free data blocks
	for ( i = 0 ; i < BLOCKSINADDR && (*address)[i] ; i++ )
	    block_free(fs->blocksys, (*address)[i]);
	
	curaddr = (*address)[NEXTADDR];
    } while ( curaddr != 0 );
    
    return 1;
}

/*
 * Write data to a file
 */
size_t file_set(filesys * fs, file * f, off_t offset, size_t size,
	const char * data)
{
    if ( offset + size >= f->size )
	if ( grow(fs, f, offset + size) != offset + size )
	    return -1;

    size_t copied = 0, copy;
    off_t fdataaddr, localoffset;
    data_block * fdata;

    while ( copied < size ) {
	if ( (fdataaddr = get_blocknmbrbyoffset(fs, f, offset + copied)) == 0 )
	    return copied;
	localoffset = (offset + copied) % sizeof(data_block);
	if ( ! (fdata = block_get(fs->blocksys, fdataaddr)) )
	    return copied;
	copy = sizeof(data_block) - localoffset;
	if ( size - copied <= copy )
	    copy = size - copied;
	memcpy((*fdata) + localoffset, data + copied, copy);
	copied += copy;
    }
    
    return copied;
}

/*
 * Read data from a file
 */
size_t file_get(filesys * fs, file * f, off_t offset, size_t size, char * data)
{
    if ( offset >= f->size )
	offset = 0;

    if ( size >= f->size - offset )
	size = f->size - offset;

    size_t copied = 0, copy;
    off_t fdataaddr, localoffset;
    data_block * fdata;

    while ( copied < size ) {
	if ( (fdataaddr = get_blocknmbrbyoffset(fs, f, offset + copied)) == 0 )
	    return copied;
	localoffset = (offset + copied) % sizeof(data_block);
	if ( ! (fdata = block_get(fs->blocksys, fdataaddr)) )
	    return copied;
	copy = sizeof(data_block) - localoffset;
	if ( size - copied <= copy )
	    copy = size - copied;
	memcpy(data + copied, (* fdata) + localoffset, copy);
	copied += copy;
    }

    return copied;
}

/*
 * Change filesize
 * (to lower or greater filesize)
 */
int file_truncate(filesys * fs, file * f, off_t size)
{
    if ( size > f->size )
	return grow(fs, f, size) == size;
    else if ( size < f->size )
	return shrink(fs, f, size) == size;
    else
	return 1;
}

/*
 * Calculates blocks used for a given filesize
 */
size_t file_size2block(size_t size)
{
    int blocks;
    blocks = size / sizeof(data_block); // data blocks
    blocks = blocks + blocks / BLOCKSINADDR;// addresses
    return blocks;
}
