#include <fs.h>

/*** Utilities ***/

/*
 * Same as strcpy but stops on '\0' or '/'
 */
static void strslashcpy(char * dst, const char * src)
{
    while ( *src != 0 && *src != '/' )
	*dst++ = *src++;
}

/*
 * Same as strlen() but stops on '\0' or '/'
 */
static int strslashlen(const char * str)
{
    int len = 0;

    while ( *str != 0 && *str++ != '/' )
	len++;

    return len;
}

/*
 * Cuts a path in an array of strings
 * last is NULL
 * (like argv)
 */
static char ** cut(const char * path)
{
    int i = 0, len;
    char ** cut = malloc(sizeof(char *));
    cut[0] = NULL;

    while ( *path != 0 ) {
	path++; // pass '/'

	cut = realloc(cut, sizeof(char *) * (i+2));

	len = strslashlen(path);
	cut[i] = malloc(len+1);
	strslashcpy(cut[i], path);
	cut[i][len] = 0;

	cut[++i] = NULL;
	path += len;
    }

    return cut;
}

/*
 * Search a file between directory entries and return its inode
 */
static int search(dir * d, char * name)
{
    int i;
    for ( i = 0 ; i < d->nb_entries ; i++ )
	if ( strcmp(d->entries[i]->filename, name) == 0 )
	    return d->entries[i]->inode;
    return -1;
}

/*
 * Returns inode for the given file path with permission checking
 */
static int resolv(filesys * fs, inodesys * is, const char * path, uid_t uid,
	gid_t gid)
{
    if ( path[1] == 0 ) // root directory
	return 0;

    inode * ino;
    int inodenb = 0;
    char ** pathcut = NULL, ** src = NULL;
    dir * d = dir_get(fs, is, 0); // Get root directory

    src = pathcut = cut(path);

    if ( d == NULL ) {
	fprintf(stderr, "resolv: Can't find root directory !");
	return -1;
    }

    while ( *pathcut != NULL ) {
	inodenb = search(d, *pathcut++);
	dir_free(d);

	if ( (ino = inode_get(is, inodenb)) == NULL )
	    return -1;

	/*
	if ( ! (ino->mode & 0001
		    || (uid == ino->uid && ino->mode & 0100)
		    || (gid == ino->gid && ino->mode & 0010) ) )
	    return -EPERM;
	*/

	if ( S_ISREG(ino->mode) ) {
	    if ( *pathcut == NULL )
		return inodenb;
	    else
		return -2;
	} else {
	    if ( (d = dir_get(fs, is, inodenb)) == NULL )
		return -1;
	}
    }

    while ( *src != NULL )
	free(*src++);

    return inodenb;
}

/*
 * Converts inode to struct data
 */
static struct stat * inode2stat(inodesys * is, int inodenb)
{
    static struct stat st;
    inode * ino;
    
    ino = inode_get(is, inodenb);
    if ( ino == NULL ) {
	return NULL;
    }

    st.st_dev = 0;
    st.st_ino = inodenb;
    st.st_mode = ino->mode;
    st.st_nlink = ino->nlink;
    st.st_uid = ino->uid;
    st.st_gid = ino->gid;
    st.st_rdev = 0;
    st.st_size = ino->size;
    st.st_blksize = sizeof(data_block);
    st.st_blocks = file_size2block(ino->size);
    st.st_atime = ino->access;
    st.st_mtime = ino->modification;
    st.st_ctime = ino->changestate;

    return & st;
}

/*** Implementation ***/

int fs_getattr(filesys * fs, inodesys * is, const char * path,
	struct stat * buf)
{
    int inodenb;
    inode * ino;
    struct stat * st;

    inodenb = resolv(fs, is, path, fuse_get_context()->uid,
	    fuse_get_context()->gid);

    if ( inodenb == -1 )
	return -ENOENT;
    else if ( inodenb == -2 )
	return -ENOTDIR;

    ino = inode_get(is, inodenb);
    if ( ino == NULL ) {
	fprintf(stderr, "Could not retrieve %d inode for directory %s\n",
		inodenb, path);
	return -ENOENT;
    }

    st = inode2stat(is, inodenb);

    if ( st == NULL )
	return -EIO;

    // memcpy(buf, st, sizeof(struct stat));
    *buf = *st;

    return 0;
}

int fs_readlink(filesys * fs, inodesys * is, const char * path, char * buf,
	size_t size)
{
    int inodenb, copied;
    file f;
    inode * ino;

    inodenb = resolv(fs, is, path, fuse_get_context()->uid,
	    fuse_get_context()->gid);
    if ( inodenb < 0 )
	return -EIO;

    ino = inode_get(is, inodenb);
    if ( ino == NULL )
	return -EIO;

    if ( ! S_ISLNK(ino->mode) )
	return -EINVAL;

    // Read
    f.faddrblck = ino->address;
    f.size = ino->size;

    copied = slink_read(fs, &f, buf, size);

    // Update inode
    if ( copied > 0 )
	ino->access = time(NULL);

    return copied;
}

int fs_symlink(filesys * fs, inodesys * is, const char * path,
	const char * target)
{
    int inodenb;
    char * name, * tmp;
    dir * parent;
    inode * ino;

    tmp = strdup(path);
    name = dirname(tmp);

    // Search parent directory
    inodenb = resolv(fs, is, name, fuse_get_context()->uid,
	    fuse_get_context()->gid);
    free(tmp);

    if ( inodenb == -1 )
	return -ENOENT;
    else if ( inodenb == -2 )
	return -ENOTDIR;

    parent = dir_get(fs, is, inodenb);
    if ( parent == NULL ) {
	fprintf(stderr, "Could not retrieve directory %s\n", path);
	return -ENOENT;
    }
    
    // Create new symlink
    inodenb = slink_create(fs, is, path, strlen(target)+1);
    if ( inodenb == -1 ) {
	dir_free(parent);
	return -ENOSPC;
    }

    tmp = strdup(path);
    name = basename(tmp);

    ino = inode_get(is, inodenb);
    if ( ino == NULL )
	return -EIO;

    // Set inode
    ino->mode = S_IFLNK | 777;
    ino->nlink = 1;
    ino->uid = fuse_get_context()->uid;
    ino->gid = fuse_get_context()->gid;
    ino->size = 0;
    ino->modification = ino->changestate = ino->access = time(NULL);

    inodenb = dir_add(parent, inodenb, name); // WARNING : inodenb stores error code

    free(tmp);
    dir_free(parent);

    switch ( inodenb ) {
	case -1: // File already exists
	    return 0;
	case 0: // Could not update directory
	    return -EIO;
	    break;
	default:
	    return 0;
    }
}

int fs_unlink(filesys * fs, inodesys * is, const char * path)
{
    int inodenb;
    inode * ino;
    file f;
    char * name, * tmp;
    dir * d;

    inodenb = resolv(fs, is, path, fuse_get_context()->uid, 
	    fuse_get_context()->gid);
    if ( inodenb < 0 )
	return -EIO;

    ino = inode_get(is, inodenb);
    if ( ino == NULL )
	return -EIO;

    if ( S_ISDIR(ino->mode) )
	return -EISDIR;

    ino->nlink--;

    if ( ino->nlink == 0 ) {
	// Remove inode
	inode_free(is, inodenb);
	// Remove data
	f.size = ino->size;
	f.faddrblck = ino->address;
	if ( ! file_free(fs, & f) )
	    return -EIO;
    }

    // Get parent directory
    tmp = strdup(path);
    name = dirname(tmp);
    inodenb = resolv(fs, is, name, fuse_get_context()->uid, 
	    fuse_get_context()->gid);
    free(tmp);
    if ( inodenb < 0 )
	return -EIO;

    d = dir_get(fs, is, inodenb);
    if ( d == NULL )
	return -EIO;

    // Remove from parent directory
    tmp = strdup(path);
    name = basename(tmp);
    if ( ! dir_remove(d, name) )
	return -EIO;
    dir_free(d);
    free(tmp);

    return 0;
}

int fs_link(filesys * fs, inodesys * is, const char * old, const char * new)
{
    int inodenb, oldinode;
    char * tmp;
    inode * ino;
    dir * d;

    // Get old
    oldinode = resolv(fs, is, old, fuse_get_context()->uid,
	    fuse_get_context()->gid);

    if ( oldinode < 0 )
	return -EIO;

    ino = inode_get(is, oldinode);
    if ( ino == NULL )
	return -EIO;

    if ( S_ISDIR(ino->mode) )
	return -EPERM;

    // Get new parent dir
    tmp = strdup(new);
    if ( tmp == NULL )
	return -ENOMEM;

    inodenb = resolv(fs, is, dirname(tmp), fuse_get_context()->uid,
	    fuse_get_context()->gid);
    free(tmp);
    if ( inodenb < 0 )
	return -ENOTDIR;

    tmp = strdup(new);
    if ( tmp == NULL )
	return -ENOMEM;

    d = dir_get(fs, is, inodenb);
    if ( d == NULL )
	return -EIO;

    // Add new
    inodenb = dir_add(d, oldinode, basename(tmp));
	    // WARNING : inodenb = error code
    free(tmp);
    dir_free(d);

    switch ( inodenb ) {
	case -1:
	    return -EEXIST;
	case 0:
	    return -EIO;
	default:
	    ino->nlink++;
	    return 0;
    }
}

int fs_rename(filesys * fs, inodesys * is, const char * old, const char * new)
{
    int ret;
    if ( (ret = fs_link(fs, is, old, new)) != 0 )
	return ret;
    if ( (ret = fs_unlink(fs, is, old)) != 0 )
	return ret;
    return 0;
}

int fs_chmod(filesys * fs, inodesys * is, const char * path, mode_t mode)
{
    int inodenb;
    inode * ino;

    inodenb = resolv(fs, is, path, fuse_get_context()->uid, 
	    fuse_get_context()->gid);
    if ( inodenb < 0 )
	return -EIO;

    ino = inode_get(is, inodenb);
    if ( ino == NULL )
	return -EIO;

    ino->mode = mode;
    ino->changestate = time(NULL);

    return 0;
}

int fs_chown(filesys * fs, inodesys * is, const char * path,
	uid_t uid, gid_t gid)
{
    int inodenb;
    inode * ino;

    inodenb = resolv(fs, is, path, fuse_get_context()->uid,
	    fuse_get_context()->gid);
    if ( inodenb < 0 )
	return -EIO;

    ino = inode_get(is, inodenb);
    if ( ino == NULL )
	return -EIO;

    ino->uid = uid;
    ino->gid = gid;
    ino->changestate = time(NULL);

    return 0;
}

int fs_truncate(filesys * fs, inodesys * is, const char * path, off_t len)
{
    int inodenb;
    file f;
    inode * ino;

    inodenb = resolv(fs, is, path, fuse_get_context()->uid,
	    fuse_get_context()->gid);
    if ( inodenb < 0 )
	return -EIO;

    ino = inode_get(is, inodenb);
    if ( ino == NULL )
	return -EIO;

    f.faddrblck = ino->address;
    f.size = ino->size;

    inodenb = file_truncate(fs, & f, len);

    ino->size = f.size;
    ino->modification = time(NULL);

    if ( inodenb )
	return 0;
    else
	return -EIO;
}

int fs_create(filesys * fs, inodesys * is, const char * path, mode_t mode)
{
    int inodenb;
    char * name, * tmp;
    dir * d;
    inode * ino;

    tmp = strdup(path);

    // Search parent directory
    inodenb = resolv(fs, is, dirname(tmp), fuse_get_context()->uid, 
	    fuse_get_context()->gid);
    free(tmp);

    if ( inodenb == -1 )
	return -ENOENT;
    else if ( inodenb == -2 )
	return -ENOTDIR;

    d = dir_get(fs, is, inodenb);
    if ( d == NULL ) {
	fprintf(stderr, "Could not retrieve directory %s\n", path);
	return -ENOENT;
    }
    
    // Create new file
    inodenb = inode_alloc(is);
    if ( inodenb == -1 ) {
	dir_free(d);
	return -ENOSPC;
    }

    tmp = strdup(path);
    name = basename(tmp);

    ino = inode_get(is, inodenb);
    if ( ino == NULL )
	return -EIO;

    file * f = file_alloc(fs);
    if ( f == NULL ) {
	inode_free(is, inodenb);
	return -ENOSPC;
    }

    // Set inode
    ino->mode = mode;
    ino->nlink = 1;
    ino->uid = fuse_get_context()->uid;
    ino->gid = fuse_get_context()->gid;
    ino->size = 0;
    ino->modification = ino->changestate = ino->access = time(NULL);
    ino->address = f->faddrblck;

    inodenb = dir_add(d, inodenb, name); // WARNING : inodenb stores error code

    free(f);
    free(tmp);
    dir_free(d);

    switch ( inodenb ) {
	case -1: // File already exists
	    if ( mode & O_EXCL )
		return -EEXIST;
	    else
		return 0;
	    break;
	case 0: // Could not update directory
	    return -ENOMEM;
	    break;
	default:
	    return 0;
    }
}

int fs_open(filesys * fs, inodesys * is, const char * path,
	struct fuse_file_info * info)
{
    int inodenb;
    inode * ino;
    
    inodenb = resolv(fs, is, path, fuse_get_context()->uid,
	    fuse_get_context()->gid);

    if ( inodenb == -1 )
	return -ENOENT;
    else if ( inodenb == -2 )
	return -ENOTDIR;

    ino = inode_get(is, inodenb);
    
    if ( ino == NULL )
	return -ENOMEM;

    if ( ! S_ISREG(ino->mode) )
	return -EISDIR;

    return 0;
}

int fs_read(filesys * fs, inodesys * is, const char * path, char * buf,
	size_t size, off_t offset)
{
    int inodenb, copied;
    file f;
    inode * ino;

    inodenb = resolv(fs, is, path, fuse_get_context()->uid,
	    fuse_get_context()->gid);
    if ( inodenb < 0 )
	return -EIO;

    ino = inode_get(is, inodenb);
    if ( ino == NULL )
	return -EIO;

    // Read
    f.faddrblck = ino->address;
    f.size = ino->size;

    copied = file_get(fs, & f, offset, size, buf);

    // Update inode
    if ( copied > 0 )
	ino->access = time(NULL);

    return copied;
}

int fs_write(filesys * fs, inodesys * is, const char * path, const char * buf,
	size_t size, off_t offset)
{
    int inodenb, copied;
    file f;
    inode * ino;
    time_t t;

    inodenb = resolv(fs, is, path, fuse_get_context()->uid,
	    fuse_get_context()->gid);
    if ( inodenb < 0 )
	return -EIO;

    ino = inode_get(is, inodenb);
    if ( ino == NULL )
	return -EIO;

    // Update file
    f.faddrblck = ino->address;
    f.size = ino->size;

    copied = file_set(fs, & f, offset, size, buf);
    if ( copied != size )
	return -EIO;

    // Update inode
    ino->size = f.size;
    t = time(NULL);
    if ( copied > 0 )
	ino->modification = t;
    ino->changestate = t;

    return copied;
}

int fs_opendir(filesys * fs, inodesys * is, const char * path,
	struct fuse_file_info * info)
{
    int inodenb;
    inode * ino;

    inodenb = resolv(fs, is, path, fuse_get_context()->uid, 
	    fuse_get_context()->gid);
    if ( inodenb < 0 )
	return -EIO;

    ino = inode_get(is, inodenb);
    if ( ino == NULL )
	return -EIO;

    if ( ! S_ISDIR(ino->mode) )
	return -ENOTDIR;

    return 0;
}

int fs_readdir(filesys * fs, inodesys * is, const char * path, void * buf,
	fuse_fill_dir_t filler)
{
    int inodenb, i;
    inode * ino;
    dir * d;
    struct stat * st;

    inodenb = resolv(fs, is, path, fuse_get_context()->uid,
	    fuse_get_context()->gid);

    if ( inodenb < 0 )
	return -ENOENT;

    ino = inode_get(is, inodenb);
    if ( ino == NULL ) {
	fprintf(stderr, "Could not retrieve %d inode for directory %s\n",
		inodenb, path);
	return -ENOENT;
    }

    d = dir_get(fs, is, inodenb);
    if ( d == NULL ) {
	fprintf(stderr, "Could not retrieve directory %s for inode %d\n", path,
		ino->address);
	return -ENOENT;
    }

    st = inode2stat(is, d->inodenb);
    filler(buf, ".", st, 0);
    filler(buf, "..", NULL, 0);

    for ( i = 0 ; i < d->nb_entries ; i++ ) {
	st = inode2stat(is, d->entries[i]->inode);
	filler(buf, d->entries[i]->filename, st, 0);
    }

    ino->access = time(NULL);

    return 0;
}

int fs_mkdir(filesys * fs, inodesys * is, const char * path, mode_t mode)
{
    int inodenb;
    char * tmp;
    dir * d, * new;

    // Search parent directory
    tmp = strdup(path);
    if ( tmp == NULL )
	return -ENOMEM;

    inodenb = resolv(fs, is, dirname(tmp), fuse_get_context()->uid,
	    fuse_get_context()->gid);
    free(tmp);
    if ( inodenb < 0 )
	return -EIO;

    d = dir_get(fs, is, inodenb);
    if ( d == NULL )
	return -EIO;

    // Create new directory
    tmp = strdup(path);
    if ( tmp == NULL ) {
	dir_free(d);
	return -ENOMEM;
    }

    new = dir_create(fs, is);   
    if ( new == NULL ) {
	dir_free(d);
	return -ENOSPC;
    }

    // Add new to parent directory
    inodenb = dir_add(d, new->inodenb, basename(tmp)); // inodenb = error code !

    switch ( inodenb ) {
	case 0:
	    dir_free(d);
	    dir_free(new);
	    return -EIO;
	case -1:
	    dir_free(d);
	    dir_free(new);
	    return -EEXIST;
	default:
	    // Set inode
	    new->ino->mode |= mode;
	    new->ino->access 
		= new->ino->modification
		= new->ino->changestate 
		= time(NULL);
	    new->ino->uid = fuse_get_context()->uid;
	    new->ino->gid = fuse_get_context()->gid;

	    d->ino->nlink++;

	    dir_free(d);
	    dir_free(new);
	    return 0;
    }
}

int fs_rmdir(filesys * fs, inodesys * is, const char * path)
{
    dir * d;
    char * tmp;
    int inodenb;

    inodenb = resolv(fs, is, path, fuse_get_context()->uid,
	    fuse_get_context()->gid);
    if ( inodenb < 0 )
	return -EIO;

    d = dir_get(fs, is, inodenb);
    if ( d == NULL )
	return -EIO;

    if ( d->nb_entries > 0 ) {
	dir_free(d);
	return -ENOTEMPTY;
    }

    if ( ! dir_delete(d) )
	return -EIO;

    dir_free(d);

    tmp = strdup(path);
    inodenb = resolv(fs, is, dirname(tmp), fuse_get_context()->uid,
	    fuse_get_context()->gid);
    free(tmp);
    if ( inodenb < 0 )
	return -EIO;

    d = dir_get(fs, is, inodenb);
    if ( d == NULL )
	return -EIO;

    tmp = strdup(path);
    if ( ! dir_remove(d, basename(tmp)) ) {
	dir_free(d);
	return -EIO;
    }

    d->ino->nlink--;

    free(tmp);
    dir_free(d);

    return 0;
}

int fs_init(char * filename, filesys ** fs, inodesys ** is, int * fsdes)
{
    *fsdes = open(filename, O_RDWR | O_CREAT, 0644);
    
    if ( *fsdes == -1 ) {
	perror("open");
	return 0;
    }

    *is = inode_init(*fsdes);
    if ( *is == NULL ) {
	fprintf(stderr, "Could not init inode system !\n");
	return 0;
    }

    *fs = file_init(*fsdes, sizeof(inodesondisk));
    if ( *fs == NULL ) {
	fprintf(stderr, "Could not init file system !\n");
	inode_destroy(*is);
	return 0;
    }

    if ( dir_init(*fs, *is) == 0 ) {
	fprintf(stderr, "Could not init filesystem directories !\n");
	inode_destroy(*is);
	file_destroy(*fs);
	return 0;
    }

    return 1;
}

void fs_destroy(filesys * fs, inodesys * is, int fsdes)
{
    inode_destroy(is);
    file_destroy(fs);

    close(fsdes);
}
