#define FUSE_USE_VERSION 25

#define _GNU_SOURCE // for getline()
#include <stdio.h>

#include <fuse/fuse.h>
#include <fuse/fuse_opt.h>
#include <errno.h>
#include <libgen.h>
#include <assert.h>

#include <fs.h>

static filesys * fs[2];
static inodesys * is[2];
static int fsdes[2];

/*** Utilities ***/

/*
 * Parse given configuration file to find file1 and file2
 */
static int parse(char * configfile, char ** file1, char ** file2)
{
    size_t linesize = 0;
    char * line = NULL;
    FILE * f = NULL;
    int linelen, linenumber = 0;

    f = fopen(configfile, "r");
    if ( f == NULL ) {
	perror("fopen");
	return 0;
    }

    while ( (linelen = getline(&line, &linesize, f)) != EOF ) {
	linenumber++;
	if ( linelen > 0 && line[0] != '#' ) {
	    if ( strncmp(line, "file1=", 5) == 0 ) {
		*file1 = strdup(line+6);
		(*file1)[strlen(*file1)-1] = 0;
	    } else if ( strncmp(line, "file2=", 5) == 0 ) {
		*file2 = strdup(line+6);
		(*file2)[strlen(*file2)-1] = 0;
	    } else
		fprintf(stderr, "Unknown value on line %d\n", linenumber);
	}
    }

    fclose(f);
    if ( line )
	free(line);

    return 1;
}

/*
 * Init all systems using given configuration file
 */
static int init(char * configfile)
{
    char * file1 = NULL, * file2 = NULL;

    if ( ! parse(configfile, & file1, & file2) ) {
	fprintf(stderr, "Unable to parse configuration file\n");
	return 0;
    }

    if ( ! fs_init(file1, fs + 0, is + 0, fsdes + 0) )
	fprintf(stderr, "Unable to initialize first filesystem\n");

    if ( ! fs_init(file2, fs + 1, is + 1, fsdes + 1) )
	fprintf(stderr, "Unable to initialize second filesystem\n");

    if ( file1 )
	free(file1);
    if ( file2 )
	free(file2);

    return 42;
}

/*
 * Repair file 0 with content of file 1
 */
static int fix(int fd0, int fd1)
{
    const int bufsize = 4096;
    char buf[bufsize];
    int len;

    if ( ftruncate(fd0, 0) != 0 ) {
	perror("[fix] ftruncate");
	return 0;
    }

    lseek(fd0, 0, SEEK_SET);
    lseek(fd1, 0, SEEK_SET);

    while ( (len = read(fd1, buf, bufsize)) != 0 ) {
	if ( len == -1 ) {
	    perror("[fix] read");
	    return 0;
	}
	if ( write(fd0, buf, len) == -1 ) {
	    perror("[fix] write");
	    return 0;
	}
    }

    return 1;
}

/*
 * Test filesystems integrity and fix if necessary
 */
static void test(int fd0, int fd1)
{
    struct stat st0, st1;

    if ( fstat(fd0, & st0) == -1 ) {
	fprintf(stderr, "Could not verify filesystem integrity !");
	perror("[test] fstat");
	return;
    }

    if ( fstat(fd1, & st1) == -1 ) {
	fprintf(stderr, "Could not verify filesystem integrity !");
	perror("[test] fstat");
	return;
    }

    // Size test
    if ( st0.st_size < st1.st_size )
	fix(fd0, fd1);
    else if ( st0.st_size > st1.st_size )
	fix(fd1, fd0);

    // Latest modification test
    if ( st0.st_mtime < st1.st_mtime )
	fix(fd0, fd1);
    else if ( st0.st_mtime > st1.st_mtime )
	fix(fd1, fd0);
}

/*** FUSE Implementation ***/

int l3_getattr(const char * path, struct stat * buf)
{
    int retfs1, retfs2;

    retfs1 = fs_getattr(fs[0], is[0], path, buf);

    if ( retfs1 == -EIO ) {
	retfs2 = fs_getattr(fs[1], is[1], path, buf);
	if ( retfs2 == -EIO )
	    return -EIO;
	fix(fsdes[0], fsdes[1]);
	return retfs2;
    }

    return retfs1;
}

int l3_unlink(const char * path)
{
    int retfs1, retfs2;

    test(fsdes[0], fsdes[1]);

    retfs1 = fs_unlink(fs[0], is[0], path);

    retfs2 = fs_unlink(fs[1], is[1], path);

    if ( retfs1 == retfs2 )
	return retfs1;

    if ( retfs2 == -EIO )
	fix(fsdes[1], fsdes[0]);
    else if ( retfs1 == -EIO )
	fix(fsdes[0], fsdes[1]);
    else
	return -EIO;

    return 0;
}

int l3_link(const char * old, const char * new)
{
    int retfs1, retfs2;

    test(fsdes[0], fsdes[1]);

    retfs1 = fs_link(fs[0], is[0], old, new);

    retfs2 = fs_link(fs[1], is[1], old, new);

    if ( retfs1 == retfs2 )
	return retfs1;

    if ( retfs2 == -EIO )
	fix(fsdes[1], fsdes[0]);
    else if ( retfs1 == -EIO )
	fix(fsdes[0], fsdes[1]);
    else
	return -EIO;

    return 0;
}

int l3_rename(const char * old, const char * new)
{
    int retfs1, retfs2;

    test(fsdes[0], fsdes[1]);

    retfs1 = fs_rename(fs[0], is[0], old, new);
    retfs2 = fs_rename(fs[1], is[1], old, new);

    if ( retfs1 == retfs2 )
	return retfs1;

    if ( retfs2 == -EIO )
	fix(fsdes[1], fsdes[0]);
    else if ( retfs1 == -EIO )
	fix(fsdes[0], fsdes[1]);
    else
	return -EIO;

    return 0;
}

int l3_chmod(const char * path, mode_t mode)
{
    int retfs1, retfs2;

    test(fsdes[0], fsdes[1]);

    retfs1 = fs_chmod(fs[0], is[0], path, mode);
    retfs2 = fs_chmod(fs[1], is[1], path, mode);

    if ( retfs1 == retfs2 )
	return retfs1;

    if ( retfs2 == -EIO )
	fix(fsdes[1], fsdes[0]);
    else if ( retfs1 == -EIO )
	fix(fsdes[0], fsdes[1]);
    else
	return -EIO;

    return 0;
}

int l3_chown(const char * path, uid_t uid, gid_t gid)
{
    int retfs1, retfs2;

    test(fsdes[0], fsdes[1]);

    retfs1 = fs_chown(fs[0], is[0], path, uid, gid);
    retfs2 = fs_chown(fs[1], is[1], path, uid, gid);

    if ( retfs1 == retfs2 )
	return retfs1;

    if ( retfs2 == -EIO )
	fix(fsdes[1], fsdes[0]);
    else if ( retfs1 == -EIO )
	fix(fsdes[0], fsdes[1]);
    else
	return -EIO;

    return 0;
}

int l3_truncate(const char * path, off_t len)
{ 
    int retfs1, retfs2;

    test(fsdes[0], fsdes[1]);

    retfs1 = fs_truncate(fs[0], is[0], path, len);
    retfs2 = fs_truncate(fs[1], is[1], path, len);

    if ( retfs1 == retfs2 )
	return retfs1;

    if ( retfs2 == -EIO )
	fix(fsdes[1], fsdes[0]);
    else if ( retfs1 == -EIO )
	fix(fsdes[0], fsdes[1]);
    else
	return -EIO;

    return 0;
}

int l3_create(const char * path, mode_t mode, struct fuse_file_info * info)
{
    (void) info;

    int retfs1, retfs2;

    test(fsdes[0], fsdes[1]);

    retfs1 = fs_create(fs[0], is[0], path, mode);
    retfs2 = fs_create(fs[1], is[1], path, mode);

    if ( retfs1 == retfs2 )
	return retfs1;

    if ( retfs2 == -EIO )
	fix(fsdes[1], fsdes[0]);
    else if ( retfs1 == -EIO )
	fix(fsdes[0], fsdes[1]);
    else
	return -EIO;

    return 0;
}

int l3_open(const char * path, struct fuse_file_info * info)
{
    int retfs1, retfs2;

    test(fsdes[0], fsdes[1]);

    retfs1 = fs_open(fs[0], is[0], path, info);

    if ( retfs1 == -EIO ) {
	retfs2 = fs_open(fs[1], is[1], path, info);
	if ( retfs2 == -EIO )
	    return -EIO;
	fix(fsdes[0], fsdes[1]);
	return retfs2;
    }

    return retfs1;
}

int l3_read(const char * path, char * buf, size_t size, off_t offset,
	struct fuse_file_info * info)
{
    (void) info;

    test(fsdes[0], fsdes[1]);

    int retfs1, retfs2;

    retfs1 = fs_read(fs[0], is[0], path, buf, size, offset);

    if ( retfs1 == -EIO ) {
	retfs2 = fs_read(fs[1], is[1], path, buf, size, offset);
	if ( retfs2 == -EIO )
	    return -EIO;
	fix(fsdes[0], fsdes[1]);
	return retfs2;
    }

    return retfs1;
}

int l3_write(const char * path, const char * buf, size_t size, off_t offset,
	struct fuse_file_info * info)
{
    (void) info;
    
    int retfs1, retfs2;

    test(fsdes[0], fsdes[1]);

    retfs1 = fs_write(fs[0], is[0], path, buf, size, offset);
    retfs2 = fs_write(fs[1], is[1], path, buf, size, offset);

    if ( retfs1 == retfs2 )
	return retfs1;

    if ( retfs2 == -EIO )
	fix(fsdes[1], fsdes[0]);
    else if ( retfs1 == -EIO )
	fix(fsdes[0], fsdes[1]);
    else
	return -EIO;

    return 0;
}

int l3_opendir(const char * path, struct fuse_file_info * info)
{
    int retfs1, retfs2;

    test(fsdes[0], fsdes[1]);

    retfs1 = fs_opendir(fs[0], is[0], path, info);

    if ( retfs1 == -EIO ) {
	retfs2 = fs_opendir(fs[1], is[1], path, info);
	if ( retfs2 == -EIO )
	    return -EIO;
	fix(fsdes[0], fsdes[1]);
	return retfs2;
    }

    return retfs1;
}

int l3_readdir(const char * path, void * buf, fuse_fill_dir_t filler, 
	off_t offset, struct fuse_file_info * info)
{
    (void) offset;
    (void) info;
    
    int retfs1, retfs2;

    test(fsdes[0], fsdes[1]);

    retfs1 = fs_readdir(fs[0], is[0], path, buf, filler);

    if ( retfs1 == -EIO ) {
	retfs2 = fs_readdir(fs[1], is[1], path, buf, filler);
	if ( retfs2 == -EIO )
	    return -EIO;
	fix(fsdes[0], fsdes[1]);
	return retfs2;
    }

    return retfs1;
}

int l3_mkdir(const char * path, mode_t mode)
{
    int retfs1, retfs2;

    test(fsdes[0], fsdes[1]);

    retfs1 = fs_mkdir(fs[0], is[0], path, mode);
    retfs2 = fs_mkdir(fs[1], is[1], path, mode);

    if ( retfs1 == retfs2 )
	return retfs1;

    if ( retfs2 == -EIO )
	fix(fsdes[1], fsdes[0]);
    else if ( retfs1 == -EIO )
	fix(fsdes[0], fsdes[1]);
    else
	return -EIO;

    return 0;
}

int l3_rmdir(const char * path)
{
    int retfs1, retfs2;

    test(fsdes[0], fsdes[1]);

    retfs1 = fs_rmdir(fs[0], is[0], path);
    retfs2 = fs_rmdir(fs[1], is[1], path);

    if ( retfs1 == retfs2 )
	return retfs1;

    if ( retfs2 == -EIO )
	fix(fsdes[1], fsdes[0]);
    else if ( retfs1 == -EIO )
	fix(fsdes[0], fsdes[1]);
    else
	return -EIO;

    return 0;
}

int l3_readlink(const char * path, char * buf, size_t size)
{
    int retfs1, retfs2;

    test(fsdes[0], fsdes[1]);

    retfs1 = fs_readlink(fs[0], is[0], path, buf, size);

    if ( retfs1 == -EIO ) {
	retfs2 = fs_readlink(fs[1], is[1], path, buf, size);
	if ( retfs2 == -EIO )
	    return -EIO;
	fix(fsdes[0], fsdes[1]);
	return retfs2;
    }

    return retfs1;
}

int l3_symlink(const char * path, const char * target)
{
    int retfs1, retfs2;

    test(fsdes[0], fsdes[1]);

    retfs1 = fs_symlink(fs[0], is[0], path, target);
    retfs2 = fs_symlink(fs[1], is[1], path, target);

    if ( retfs1 == retfs2 )
	return retfs1;

    if ( retfs2 == -EIO )
	fix(fsdes[1], fsdes[0]);
    else if ( retfs1 == -EIO )
	fix(fsdes[0], fsdes[1]);
    else
	return -EIO;

    return 0;
}

static struct fuse_operations l3_op = {
    .getattr = l3_getattr,
    .readlink = l3_readlink,
    .mkdir = l3_mkdir,
    .unlink = l3_unlink,
    .rmdir = l3_rmdir,
    .symlink = l3_symlink,
    .rename = l3_rename,
    .link = l3_link,
    .chmod = l3_chmod,
    .chown = l3_chown,
    .truncate = l3_truncate,
    .open = l3_open,
    .read = l3_read,
    .write = l3_write,
    .opendir = l3_opendir,
    .readdir = l3_readdir,
    .create = l3_create,
};

int main(int argc, char ** argv)
{
    if ( argc < 3 ) {
	fprintf(stderr, "Usage : %s <configfile> <mountpoint> "
		"[FUSE options...]\n", argv[0]);
	exit(EXIT_FAILURE);
    }

    if ( ! init(argv[1]) )
	exit(EXIT_FAILURE);

    if ( fuse_main(argc - 1, argv + 1, & l3_op) == 0 )
	exit(EXIT_SUCCESS);
    else
	exit(EXIT_FAILURE);
}
